from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import *
import http.client
from typing import NamedTuple
from unittest.case import TestCase

from flask import json

from zsl import Zsl, inject
from zsl.application.containers.web_container import WebContainer
from zsl.application.error_handler import DefaultErrorHandler, ErrorResponse
from zsl.application.modules.web.cors import CORSConfiguration
from zsl.db.helpers.pagination import PaginationRequest
from zsl.interface.task import create_simple_model, fill_model_with_payload, payload_into_model
from zsl.interface.web.performers.task import perform_web_task
from zsl.interface.web.utils.execution import execute_web_task
from zsl.router.task import TaskConfiguration
from zsl.task.job_context import WebJobContext
from zsl.task.task_decorator import json_output
from zsl.testing.db import IN_MEMORY_DB_SETTINGS
from zsl.testing.http import HTTPTestCase
from zsl.testing.zsl import ZslTestCase, ZslTestConfiguration


class PaginationParsingTaskResult(object):

    stuffage: str = ""
    page_no: int = 0
    page_size: int = 0


PaginationParsingRequest = create_simple_model(
    'PaginationParsingRequest',
    ['stuffage', 'pagination'],
    {'pagination': PaginationRequest()}
)

_STUFFAGE = "Hello world!"
_PAGE_NO = 6
_PAGE_SIZE = 65


class PaginationParsingTask(object):

    @json_output
    @payload_into_model(PaginationParsingRequest)
    def perform(self, request: PaginationParsingRequest):
        result = PaginationParsingTaskResult()
        result.stuffage = request.stuffage
        result.page_no = request.pagination.page_no
        result.page_size = request.pagination.page_size
        return result


class PaginationParsingTestCase(ZslTestCase, HTTPTestCase, TestCase):
    ZSL_TEST_CONFIGURATION = ZslTestConfiguration(
        app_name='PaginationParsingTestCase',
        container=WebContainer,
        config_object={
            "CORS": CORSConfiguration(origin='origin'),
            "TASKS": TaskConfiguration()
            .create_namespace('n')
            .add_routes({'p': PaginationParsingTask})
            .get_configuration()
        }
    )

    @inject(app=Zsl)
    def testErrorTaskExecution(self, app):
        with self.getHTTPClient() as client:
            random_token = '1'

            response = self.requestTask(
                client,
                '/n/p',
                {
                    "stuffage": _STUFFAGE,
                    "pagination": {
                        "page_no": _PAGE_NO,
                        "page_size": _PAGE_SIZE
                    }
                }
            )

            parsing_response = PaginationParsingTaskResult()
            fill_model_with_payload(json.loads(response.data), parsing_response)

            self.assertEqual(_STUFFAGE, parsing_response.stuffage)
            self.assertEqual(_PAGE_NO, parsing_response.page_no)
            self.assertEqual(_PAGE_SIZE, parsing_response.page_size)



__author__ = 'julius'

from unittest.case import TestCase

from werkzeug.datastructures import ImmutableMultiDict

from zsl.utils.request_helper import args_to_dict


class InflectionTestCase(TestCase):

    def test_empty(self):
        args = ImmutableMultiDict([])
        args_dict = args_to_dict(args)
        self.assertEqual(args_dict, {}, "Arguments dictionary should be empty for empty arguments object")

    def test_one_param(self):
        args = ImmutableMultiDict([('a', 1)])
        args_dict = args_to_dict(args)
        self.assertEqual(args_dict, {'a': 1}, "Arguments dictionary should contain the one argument")

    def test_more_params(self):
        args = ImmutableMultiDict([('a', 1), ('b', 2), ('c', 42)])
        args_dict = args_to_dict(args)
        self.assertEqual(args_dict, {'a': 1, 'b': 2, 'c': 42}, "Arguments dictionary should contain the same "
                                                               "parameters as the arguments object")

    def test_multi_param(self):
        args = ImmutableMultiDict([('a', 2), ('a', 4), ('a', 8)])
        args_dict = args_to_dict(args)
        self.assertEqual(args_dict, {'a': [2, 4, 8]}, "The param should be list of all of its values")

    def test_multi_params_with_other_params(self):
        args = ImmutableMultiDict([('a', 3), ('b', 'PWR UP'), ('a', 9), ('a', 27)])
        args_dict = args_to_dict(args)
        self.assertEqual(args_dict, {'a': [3, 9, 27], 'b': 'PWR UP'}, "Arguments dictionary should contain the same "
                                                                      "parameters as the arguments object")



"""
:mod:`zsl.contrib.modules.sentry_module`
----------------------------------------

Sentry module is responsible for handling errors and sending them to a sentry
backend. Just provide a DSN (url) of the backend and you are all set up.

To add the Sentry support just add `SentryModule` to your container.

.. code-block:: python

    class MyApplicationContainer(WebContainer):
        sentry = SentryModule

Configuration
~~~~~~~~~~~~~

To configure one needs to provide a valid `SentryConfiguration` usually in
`default_settings.py` or in your profiles.

.. code-block:: python

    SENTRY = SentryConfiguration(dsn="...")

The following options are available:

    1. `dsn`: DSN/URl of your Sentry project,
    2. `environment`: the environment (release defaults to the version of your application),
    3. `tags`: additional tags str to str dictionary,
    4. `register_sentry_logging_handler`: bool value indicating if logging handler should be added to logging,
    5. `sentry_handler_logging_level`: minimal log level of log messages sent to sentry logging handler.

To test the functionality use CLI

.. code-block:: bash

    python app.py sentry test

"""
from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import *
import logging

from injector import Binder, provides, singleton
from sentry_sdk import HttpTransport
from sentry_sdk.integrations.logging import LoggingIntegration

from zsl import Config, Module, Zsl, inject
from zsl.application.error_handler import register
from zsl.application.modules.cli_module import ZslCli
from zsl.contrib.sentry.sentry_config import SentryConfiguration
from zsl.errors import ErrorProcessor, ZslError
from zsl.utils.injection_helper import simple_bind

try:
    import sentry_sdk
except ImportError:
    CommandLine = None
    logging.getLogger(__name__).exception(
        "Can not import sentry sdk. Please install it first `pip install zsl [sentry-sdk]`.")
    raise


class SentryCli(object):
    pass

    """Sentry CLI interface support."""

    @inject(zsl_cli=ZslCli)
    def __init__(self, zsl_cli):
        # type: (ZslCli) -> None
        logging.getLogger(__name__).debug("Creating Sentry CLI.")

        @zsl_cli.cli.group()
        def sentry():
            pass

        @sentry.command(help='Send a test error to the Sentry backend.')
        def test():
            print('Sending test Sentry message.')
            raise ZslError("Sentry test error from Zsl.")

        self._sentry = sentry

    @property
    def sentry(self):
        return self._sentry


class SentryErrorProcessor(ErrorProcessor):

    @inject(config=SentryConfiguration, zsl=Zsl)
    def __init__(self, config, zsl):
        # type: (SentryConfiguration, Zsl)->None
        self._init_sdk(config, zsl)

    @staticmethod
    def _init_sdk(config, zsl):
        # type: (SentryConfiguration, Zsl)->None
        logging_integration = SentryErrorProcessor._register_logging_handler(config)

        sentry_sdk.init(
            dsn=config.dsn,
            transport=HttpTransport,
            environment=config.environment,
            release=zsl.get_version(),
            integrations=[logging_integration] if logging_integration else None,
        )

        for key, value in config.tags.items():
            sentry_sdk.set_tag(key, value)

    @staticmethod
    def _register_logging_handler(config):
        # type: (SentryConfiguration)->LoggingIntegration
        return LoggingIntegration(
            level=None,
            event_level=config.sentry_logging_handler_level if config.register_logging_handler else None,
        )

    def handle(self, e):
        logging.getLogger(__name__).info('Sending error message for {0}.'.format(e))
        sentry_sdk.capture_exception(e)


class SentryModule(Module):
    """Adds Sentry support."""

    SENTRY_CONFIG_NAME = 'SENTRY'

    @provides(SentryConfiguration)
    @inject(config=Config)
    def provide_sentry_configuration(self, config):
        # type: (Config) -> SentryConfiguration
        return config.get(SentryModule.SENTRY_CONFIG_NAME)

    def configure(self, binder):
        # type: (Binder) -> None
        simple_bind(binder, SentryCli, singleton)
        register(SentryErrorProcessor())



"""
:mod:`zsl.service.service`
--------------------------

.. moduleauthor:: Martin Babka <babka@atteq.com>,
                  Peter Morihladko <morihladko@atteq.com>
"""
from __future__ import annotations

from contextlib import contextmanager
from functools import wraps
import logging
from typing import Callable, Generator

from sqlalchemy.engine.base import Engine
from sqlalchemy.orm.session import Session

from zsl import Injected, Zsl, inject
from zsl.application.modules.alchemy_module import EmptyTransactionalHolder, SessionFactory, TransactionHolderFactory

_EMPTY_TX_HOLDER = EmptyTransactionalHolder()

logger = logging.getLogger(__name__)


class TransactionalSupportMixin:
    """This mixin allows the objects to access transactional holder."""

    @property
    def _orm(self) -> Session:
        return getattr(self, _TX_HOLDER_ATTRIBUTE, _EMPTY_TX_HOLDER).session


class Service(TransactionalSupportMixin):
    """
    Main service class.
    """

    @inject(app=Zsl, engine=Engine)
    def __init__(self, app=Injected, engine=Injected):
        """Constructor - initializes and injects the needed libraries."""
        super(Service, self).__init__()
        self._app = app
        self._engine = engine


_TX_HOLDER_ATTRIBUTE = "_tx_holder"


@inject(session_factory=SessionFactory)
def _get_session_factory(session_factory: SessionFactory = Injected) -> SessionFactory:
    return session_factory


@inject(tx_holder_factory=TransactionHolderFactory)
def _get_tx_holder_factory(
    tx_holder_factory: TransactionHolderFactory = Injected,
) -> TransactionHolderFactory:
    return tx_holder_factory


@contextmanager
def tx_session(service: TransactionalSupportMixin) -> Generator[Session, None, None]:
    """
    Context manager for transactional session.

    :param service: Service instance.
    :return: Session.

    .. code-block:: python

        from zsl.service import tx_session

        class MyService(Service):
            def my_method(self):
                with tx_session(self) as session:
                    self._repository.create(session, 2, 3, 4)

    """
    session_factory = _get_session_factory()
    transaction_is_open = False

    assert isinstance(service, TransactionalSupportMixin)

    if hasattr(service, _TX_HOLDER_ATTRIBUTE):
        tx_holder = getattr(service, _TX_HOLDER_ATTRIBUTE)
        assert tx_holder.has_session()
    else:
        tx_holder = _get_tx_holder_factory().create_transaction_holder()
        setattr(service, _TX_HOLDER_ATTRIBUTE, tx_holder)
        transaction_is_open = True

    try:
        logger.debug("Entering transactional method.")
        if not tx_holder.has_session():
            session = session_factory.create_session()
            tx_holder.inject_session(session)

        tx_holder.begin()

        yield tx_holder.session

        if transaction_is_open:
            tx_holder.run_transaction_callbacks()
            tx_holder.commit()

    except Exception:
        if transaction_is_open:
            tx_holder.rollback()
        raise
    finally:
        logger.debug("Finishing transactional method.")
        if transaction_is_open:
            tx_holder.close()
            delattr(service, _TX_HOLDER_ATTRIBUTE)
            logger.debug("Closing TX session.")


def transactional(f: Callable) -> Callable:
    """
    Decorator for making a function transactional.

    :param f: Function to decorate.
    :return: Decorated function.

    .. code-block:: python

        from zsl.service import transactional

        class MyService(Service):
            @transactional
            def my_function(self):
                self._repository.create(self._orm, 2, 3, 4)
    """

    @wraps(f)
    def transactional_f(service, *args, **kwargs):
        with tx_session(service):
            return f(service, *args, **kwargs)

    return transactional_f



import http.client
from unittest.case import TestCase

from zsl import Zsl, inject
from zsl.application.containers.core_container import CoreContainer
from zsl.router.task import TaskConfiguration
from zsl.testing.db import IN_MEMORY_DB_SETTINGS
from zsl.testing.zsl import ZslTestCase, ZslTestConfiguration


class AppVersionTestCase(ZslTestCase, TestCase):
    CONFIG = IN_MEMORY_DB_SETTINGS.copy()
    CONFIG.update(
        TASKS=TaskConfiguration()
    )
    ZSL_TEST_CONFIGURATION = ZslTestConfiguration(
        app_name='ErrorHandlingTestCase', container=CoreContainer,
        version="1234.1111.4321",
        config_object=CONFIG)

    @inject(app=Zsl)
    def testErrorTaskExecution(self, app: Zsl) -> None:
        self.assertEqual(Zsl.VERSION, app.zsl_version)
        self.assertEqual("1234.1111.4321", app.app_version)
        self.assertEqual(Zsl.VERSION + ":1234.1111.4321", app.get_version())
        self.assertEqual(Zsl.VERSION + ":1234.1111.4321", app.version)


class NoAppVersionTestCase(ZslTestCase, TestCase):
    CONFIG = IN_MEMORY_DB_SETTINGS.copy()
    CONFIG.update(
        TASKS=TaskConfiguration()
    )
    ZSL_TEST_CONFIGURATION = ZslTestConfiguration(
        app_name='ErrorHandlingTestCase', container=CoreContainer,
        config_object=CONFIG)

    @inject(app=Zsl)
    def testErrorTaskExecution(self, app: Zsl) -> None:
        self.assertEqual(Zsl.VERSION, app.zsl_version)
        self.assertIsNone(app.app_version)
        self.assertEqual(Zsl.VERSION, app.get_version())
        self.assertEqual(Zsl.VERSION, app.version)



"""
:mod:`zsl.utils.request_helper`

Helper module for working with request.

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals

from future.utils import viewitems


def args_to_dict(args):
    """Converts request arguments to a simple dictionary. If an argument
    appears multiple times, the resulting dictionary will contain list
    of all of its values.

    :param args: flask's request args object
    :type args: request.args
    :return: a simple dict from args
    :rtype: dict
    """
    args_dict = {}
    for param_name, param_value in args.items(multi=True):
        if param_name not in args_dict:
            args_dict[param_name] = param_value
        elif not isinstance(args_dict[param_name], list):
            current_value = args_dict[param_name]
            args_dict[param_name] = [current_value, param_value]
        else:
            args_dict[param_name].append(param_value)

    return args_dict
